/*************************************************************************

      Graphics Research Group - Department of Computer Science
                  University of Sheffield, UK
                      Michael Meredith
                     Copyright (c) 2000
                    All Rights Reserved.


  Permission is hereby granted, free of charge, to use this software 
  and its documentation without restriction, including without limitation 
  the rights to use, copy, modify, merge and publish, subject to the following 
  conditions:

   1. The code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Any modifications must be clearly marked as such.

   3. Original authors' names are not deleted.

   4. The authors' names are not used to endorse or promote products
      derived from this software without specific prior written
      permission from the author.

   5. This software is used in a research, non-profit capacity only and not 
      for commercial or profit applications unless specific prior written
      permission from the author is given.


  THE UNIVERSITY OF SHEFFIELD AND THE CONTRIBUTORS TO THIS WORK
  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
  SHALL THE UNIVERSITY OF SHEFFIELD NOR THE CONTRIBUTORS BE LIABLE
  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
  THIS SOFTWARE.

*************************************************************************/

#include "DisplayGL.h"

unsigned char threeto8[8] = {0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377};
unsigned char twoto8[4] = {0, 0x55, 0xaa, 0xff};
unsigned char oneto8[2] = {0, 255};
static int defaultOverride[13] = {0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91};
static PALETTEENTRY defaultPalEntry[20] = 
{
	{ 0,   0,   0,    0 },
	{ 0x80,0,   0,    0 },
	{ 0,   0x80,0,    0 },
	{ 0x80,0x80,0,    0 },
	{ 0,   0,   0x80, 0 },
	{ 0x80,0,   0x80, 0 },
	{ 0,   0x80,0x80, 0 },
	{ 0xC0,0xC0,0xC0, 0 },

	{ 192, 220, 192,  0 },
	{ 166, 202, 240,  0 },
	{ 255, 251, 240,  0 },
	{ 160, 160, 164,  0 },

	{ 0x80,0x80,0x80, 0 },
	{ 0xFF,0,   0,    0 },
	{ 0,   0xFF,0,    0 },
	{ 0xFF,0xFF,0,    0 },
	{ 0,   0,   0xFF, 0 },
	{ 0xFF,0,   0xFF, 0 },
	{ 0,   0xFF,0xFF, 0 },
	{ 0xFF,0xFF,0xFF, 0 }
};

BEGIN_MESSAGE_MAP(DisplayGL, CWnd)
  ON_WM_SIZE()
  ON_WM_PAINT()
  ON_WM_CREATE()
  ON_WM_LBUTTONDOWN()
  ON_WM_RBUTTONDOWN()
  ON_WM_MOUSEMOVE()
END_MESSAGE_MAP()

DisplayGL::DisplayGL() : body(0), header(0), m_pDC(0), lasttime(::GetTickCount()), frames(0)
{
  fps[0] = '\0';
  centre[0] = centre[1] = centre[2] = 0.0f;
}

DisplayGL::~DisplayGL()
{
  wglDeleteContext(hrc);
}

int DisplayGL::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
  Init();
  return 0;
}

void DisplayGL::Init()
{
  PIXELFORMATDESCRIPTOR pfd;
  int  n;
	HGLRC	hrc;
	GLfloat	fAspect;

  m_pDC = new CClientDC(this);

  ASSERT(m_pDC != NULL);

  if (!bSetupPixelFormat())
    return;

  n = ::GetPixelFormat(m_pDC->GetSafeHdc());
  ::DescribePixelFormat(m_pDC->GetSafeHdc(), n, sizeof(pfd), &pfd);

  CreateRGBPalette();

  hrc = wglCreateContext(m_pDC->GetSafeHdc());
  wglMakeCurrent(m_pDC->GetSafeHdc(), hrc);

  glClearDepth(1.0f);
  glEnable(GL_DEPTH_TEST);
  
  CRect rect;
  GetClientRect(rect);
  if (rect.bottom)
	  fAspect = (GLfloat)rect.right/rect.bottom;
	else
	  fAspect = 1.0f;
	
  glViewport(0, 0, rect.right, rect.bottom);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(40.0f, fAspect, 0.0f, 10.0f);
  glMatrixMode(GL_MODELVIEW);
}

BOOL DisplayGL::bSetupPixelFormat()
{
  static PIXELFORMATDESCRIPTOR pfd = 
	{
    sizeof(PIXELFORMATDESCRIPTOR), 1, PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER, PFD_TYPE_RGBA, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, PFD_MAIN_PLANE, 0, 0, 0, 0};
  int pixelformat;

  if ((pixelformat = ChoosePixelFormat(m_pDC->GetSafeHdc(), &pfd)) == 0)
  {
    MessageBox("ChoosePixelFormat failed");
    return FALSE;
  }

  if (SetPixelFormat(m_pDC->GetSafeHdc(), pixelformat, &pfd) == FALSE)
  {
    MessageBox("SetPixelFormat failed");
    return FALSE;
  }
  return TRUE;
}

void DisplayGL::CreateRGBPalette()
{
  PIXELFORMATDESCRIPTOR pfd;
  LOGPALETTE *pPal;
  int n, i;
 
  n = ::GetPixelFormat(m_pDC->GetSafeHdc());
  ::DescribePixelFormat(m_pDC->GetSafeHdc(), n, sizeof(pfd), &pfd);

  if (pfd.dwFlags & PFD_NEED_PALETTE)
  {
    n = 1 << pfd.cColorBits;
    pPal = (PLOGPALETTE) new char[sizeof(LOGPALETTE) + n * sizeof(PALETTEENTRY)];

    ASSERT(pPal != NULL);

    pPal->palVersion = 0x300;
    pPal->palNumEntries = n;
    for (i=0; i<n; i++)
    {
      pPal->palPalEntry[i].peRed = ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
      pPal->palPalEntry[i].peGreen = ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
      pPal->palPalEntry[i].peBlue = ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
      pPal->palPalEntry[i].peFlags = 0;
    }

    /* fix up the palette to include the default GDI palette */
    if ((pfd.cColorBits == 8) && (pfd.cRedBits == 3) && (pfd.cRedShift == 0) && (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) && (pfd.cBlueBits == 2) && (pfd.cBlueShift == 6))
    {
			for (i = 1 ; i <= 12 ; i++)
        pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
    }

    m_cPalette.CreatePalette(pPal);
    delete pPal;

    m_pOldPalette = m_pDC->SelectPalette(&m_cPalette, FALSE);
    m_pDC->RealizePalette();
  }
}

unsigned char DisplayGL::ComponentFromIndex(int i, UINT nbits, UINT shift)
{
  unsigned char val;

  val = (unsigned char) (i >> shift);
  switch (nbits) 
	{
  case 1:
    val &= 0x1;
    return oneto8[val];
  case 2:
     val &= 0x3;
     return twoto8[val];
  case 3:
     val &= 0x7;
     return threeto8[val];
  default:
     return 0;
  }
}

void DisplayGL::OnPaint()
{
	CPaintDC dc(this);
  Draw();
}

void DisplayGL::OnSize(UINT nType, int cx, int cy)
{
  cx = (cx==0) ? 1 : cx;
  cy = (cy==0) ? 1 : cy;
  glViewport(0, 0, cx, cy);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(40.0f, (GLfloat)cx/cy, 1.0f, 200.0f);
  glMatrixMode(GL_MODELVIEW);
}

void DisplayGL::Draw()
{
  static BOOL bBusy = FALSE;
  if(bBusy || !body) 
	  return;
  bBusy = TRUE;

  glClearColor(1.0f,  1.0f,  1.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  glShadeModel(GL_SMOOTH);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

    glTranslatef(0.0f, 0.0f, -7.0f);
    glRotatef(centre[0], 1.0f, 0.0f, 0.0f);
    glRotatef(centre[1], 0.0f, 1.0f, 0.0f);
    glRotatef(centre[2], 0.0f, 0.0f, 1.0f);

    if (body)
      EvaluateChildren(body);

  glPopMatrix();
  glFinish();
  SwapBuffers(wglGetCurrentDC());

  DWORD time = ::GetTickCount();
  if (lasttime < time - 2000)
  {
    _gcvt(frames * 1000/ (float) (time - lasttime), 6, fps);
    frames = 0;
    lasttime = time;
    char buff[40];
    strcpy(buff, "DisplayGL - ");
    strcat(buff, fps);
    SetWindowText(buff);
  }
  else
    ++frames;

  bBusy = FALSE;
}

void DisplayGL::SetupSegments(NODE* segs, MOCAPHEADER *header)
{
  this->header = header;
  body = segs;
}

void DisplayGL::EvaluateChildren(NODE* node)
{
  glPushMatrix();
    glTranslatef(node->offset[0] + node->froset[header->currentframe][0], node->offset[1] + node->froset[header->currentframe][1], node->offset[2] + node->froset[header->currentframe][2]);

    glRotatef(node->euler[0], (float) header->euler[0][0], (float) header->euler[0][1], (float) header->euler[0][2]);
    glRotatef(node->euler[1], (float) header->euler[1][0], (float) header->euler[1][1], (float) header->euler[1][2]);
    glRotatef(node->euler[2], (float) header->euler[2][0], (float) header->euler[2][1], (float) header->euler[2][2]);

    glRotatef(node->freuler[header->currentframe][0], (float) header->euler[0][0], (float) header->euler[0][1], (float) header->euler[0][2]);
    glRotatef(node->freuler[header->currentframe][1], (float) header->euler[1][0], (float) header->euler[1][1], (float) header->euler[1][2]);
    glRotatef(node->freuler[header->currentframe][2], (float) header->euler[2][0], (float) header->euler[2][1], (float) header->euler[2][2]);

    glBegin(GL_LINES);
      glColor3f(node->colour[0], node->colour[1], node->colour[2]);
      glVertex3f(0.0f, 0.0f, 0.0f);
      glVertex3f(node->length[0] * node->scale[header->currentframe], node->length[1] * node->scale[header->currentframe], node->length[2] * node->scale[header->currentframe]);
    glEnd();

    if (node->children)
    {
      for (int i=0; i<node->noofchildren; i++)
        EvaluateChildren(node->children[i]);
    }

  glPopMatrix();
}

void DisplayGL::OnLButtonDown(UINT nFlags, CPoint point)
{
  last = point;
}
void DisplayGL::OnRButtonDown(UINT nFlags, CPoint point)
{
  last = point;
}

void DisplayGL::OnMouseMove(UINT nFlags, CPoint point)
{
  if (nFlags & MK_LBUTTON)
  {
    centre[1]+=(point.x - last.x);
    centre[0]+=(point.y - last.y);
  }
  else if (nFlags & MK_RBUTTON)
  {
    centre[2]-=(point.x - last.x);
  }
  last = point;
}
