/*************************************************************************

      Graphics Research Group - Department of Computer Science
                  University of Sheffield, UK
                      Michael Meredith
                     Copyright (c) 2000
                    All Rights Reserved.


  Permission is hereby granted, free of charge, to use this software 
  and its documentation without restriction, including without limitation 
  the rights to use, copy, modify, merge and publish, subject to the following 
  conditions:

   1. The code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Any modifications must be clearly marked as such.

   3. Original authors' names are not deleted.

   4. The authors' names are not used to endorse or promote products
      derived from this software without specific prior written
      permission from the author.

   5. This software is used in a research, non-profit capacity only and not 
      for commercial or profit applications unless specific prior written
      permission from the author is given.


  THE UNIVERSITY OF SHEFFIELD AND THE CONTRIBUTORS TO THIS WORK
  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
  SHALL THE UNIVERSITY OF SHEFFIELD NOR THE CONTRIBUTORS BE LIABLE
  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
  THIS SOFTWARE.

*************************************************************************/

#include "Main.h"
#include "Resource.h"
#include "HTRFormat.h"
#include "BVHFormat.h"

BEGIN_MESSAGE_MAP(MainWnd, CWnd)
  ON_WM_PAINT()
  ON_WM_CREATE()
  ON_WM_DESTROY()
  ON_WM_TIMER()
END_MESSAGE_MAP()

Main theApp;

BOOL Main::InitInstance()
{
  m_pMainWnd = new MainWnd(m_lpCmdLine);
  return TRUE;
}

BOOL Main::OnIdle(LONG lCount)
{
  PostMessage(m_pMainWnd->m_hWnd, WM_TIMER, 0, 0);
  return true;
}

MainWnd::MainWnd(const char* m_lpCmdLine)
{
  char *pathname = (char*) malloc(30);
  strcpy(pathname, "Data\\");

  int left = (GetSystemMetrics(SM_CXFULLSCREEN) - 800) /  2;
  int top = (GetSystemMetrics(SM_CYFULLSCREEN) - 600) /2;
  if (left < 0)
    left = 0;
  if (top < 0)
    top = 0;
  CRect rect(10, 10, 200, 200);
  LOGBRUSH brush = {BS_SOLID, 0xC0C0C0};
	HBRUSH bkgrd = ::CreateBrushIndirect(&brush);
  const char* classname = AfxRegisterWndClass(CS_DBLCLKS|CS_HREDRAW|CS_VREDRAW|CS_SAVEBITS, ::LoadCursor(NULL, IDC_ARROW), bkgrd, ::LoadIcon(AfxGetResourceHandle(), MAKEINTRESOURCE(IDI_LOGO)));

  CRect rect2(left, top, left + 800, top + 600);
  Output3D = new DisplayGL();
  Output3D->CreateEx(NULL, classname, "DisplayGL", WS_OVERLAPPEDWINDOW, rect2, this , NULL, NULL);
  CreateEx(WS_EX_CONTROLPARENT, classname, "MainWnd", WS_CLIPCHILDREN | WS_OVERLAPPEDWINDOW, rect, Output3D, NULL, NULL);

  // Setup the basic body structure ready for the system to build on it
  char mocapfilename[200];
  if (m_lpCmdLine[0])
    strcpy(mocapfilename,m_lpCmdLine);  // get file to load in from the command line parameters
  else  // no specified file so use a default one
  {
    strcpy(mocapfilename, pathname);
//    strcat(mocapfilename, "goal.htr");
//    strcat(mocapfilename, "catch.htr");
    strcat(mocapfilename, "dance01.bvh");
//    strcat(mocapfilename, "vogue.bvh");
  }
  
  MocapData* mocapfile;
  if (strstrEx(mocapfilename, ".HTR")!=-1)
    mocapfile = new HTRFormat(&header);
  else if (strstrEx(mocapfilename, ".BVH")!=-1)
    mocapfile = new BVHFormat(&header);
  else
    mocapfile = 0;


  if (!mocapfile)
  {
    MessageBox("Unrecognised file extension.");
    exit(1);
  }
  else if (!mocapfile->ImportData(mocapfilename))
  {
    MessageBox(mocapfile->GetError());
    exit(1);
  }
  else
  {
    body = mocapfile->GetRootNode();
    nodelist = mocapfile->GetNodeList();
    header.currentframe = 0;

    Output3D->SetupSegments(body, &header);
    Output3D->ShowWindow(SW_SHOW);
  }
}

MainWnd::~MainWnd()
{}

int MainWnd::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
  return 0;
}

void MainWnd::OnDestroy()
{
  for (int i=0; i<header.noofsegments; ++i)
  {
    for (long j=0; j<header.noofframes; ++j)
    {
      free(nodelist[i]->froset[j]);
      free(nodelist[i]->freuler[j]);
    }
    free(nodelist[i]->scale);
    free(nodelist[i]->froset);
    free(nodelist[i]->freuler);
    free(nodelist[i]->children);
    free(nodelist[i]->name);
    free(nodelist[i]);
  }
  free(nodelist);
}

void MainWnd::OnPaint()
{

}

void MainWnd::OnTimer(UINT nIDEvent)
{
  static long starttime = GetTickCount();
  header.currentframe = ((GetTickCount() - starttime) * header.datarate / 1000) % header.noofframes;

  Output3D->Draw();
}

void MainWnd::OutputNodeList()
{
  FILE *output = fopen("C:\\Debug.txt", "wb");
  char buffer[20];
  char nline[2];
  nline[0] = '\r';
  nline[1] = '\n';
  for (int j=0; j<header.noofsegments; ++j)
  {
    fwrite("Name: ", 6,1,output);
    fwrite(nodelist[j]->name,strlen(nodelist[j]->name),1,output);
    fwrite(nline,2,1,output);
    fwrite("Length X: ", 10, 1, output);
    gcvt(nodelist[j]->length[0], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);
    fwrite(nline,2,1,output);
    fwrite("Length Y: ", 10, 1, output);
    gcvt(nodelist[j]->length[1], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);
    fwrite(nline,2,1,output);
    fwrite("Length Z: ", 10, 1, output);
    gcvt(nodelist[j]->length[2], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);

    fwrite(nline,2,1,output);
    fwrite("Offest X: ", 10, 1, output);
    gcvt(nodelist[j]->offset[0], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);
    fwrite(nline,2,1,output);
    fwrite("Offest Y: ", 10, 1, output);
    gcvt(nodelist[j]->offset[1], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);
    fwrite(nline,2,1,output);
    fwrite("Offest Z: ", 10, 1, output);
    gcvt(nodelist[j]->offset[2], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);

    fwrite(nline,2,1,output);
    fwrite("Euler X: ", 9, 1, output);
    gcvt(nodelist[j]->euler[0], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);
    fwrite(nline,2,1,output);
    fwrite("Euler Y: ", 9, 1, output);
    gcvt(nodelist[j]->euler[1], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);
    fwrite(nline,2,1,output);
    fwrite("Euler Z: ", 9, 1, output);
    gcvt(nodelist[j]->euler[2], 6, buffer);
    fwrite(buffer,strlen(buffer),1,output);

    fwrite(nline,2,1,output);
    fwrite("Children: ", 10, 1, output);
    _itoa(nodelist[j]->noofchildren,buffer,10);
    fwrite(buffer, strlen(buffer), 1, output);
    fwrite(nline,2,1,output);
    fwrite(nline,2,1,output);
  }
  fclose(output);
}